import numpy as np
import re
from dataclasses import dataclass

from elphem.lattice.cell import Cell3D, Cell2D, Cell1D
from elphem.lattice.primitive_cell import PrimitiveCell3D, PrimitiveCell2D, PrimitiveCell1D
from elphem.lattice.lattice_constant import LatticeConstant3D, LatticeConstant2D, LatticeConstant1D
from elphem.common.brillouin import SpecialPoints3D, SpecialPoints2D, SpecialPoints1D
from elphem.lattice.path import PathValues

class ReciprocalCell:
    """A class for reciprocal cells
    Attributes:
        n_dim (int): Number of dimensions
    """
    def __init__(self):
        self.n_dim = None
    
    @staticmethod
    def split_fractional_k_name(fractional_k_name: str) -> tuple:
        """Split fractional special k points.

        Args:
            fractional_k_name (str): A string of fractional special k points

        Raises:
            ValueError: If the string format is invalid.

        Returns:
            tuple: fraction and the name of a special k point.
        """
        match = re.match(r"([0-9.]*)([A-Z]+)", fractional_k_name)
        
        if match:
            fraction_str = match.group(1)
            k_name = match.group(2)

            number = float(fraction_str) if fraction_str else 1.0
            
            return number, k_name
        else:
            raise ValueError("Invalid input string format. Ensure the input contains only a combination of numbers and k-names.")

    def get_path(self, fractional_k_names: list[str], n_split: int) -> PathValues:
        """Calculates a path through specified special points in the Brillouin zone.

        Args:
            fractional_k_names (list[str]): List of special point names to form the path.
            n_split (int): Number of points between each special point.

        Returns:
            k_path (PathValue): Values with path.
        """
        fractional_k_names_tuple = [self.split_fractional_k_name(fractional_k_name) for fractional_k_name in fractional_k_names]
        
        k_via = [fractional_k_name[0] * self.calculate_special_k(fractional_k_name[1]) for fractional_k_name in fractional_k_names_tuple]
        n_via = len(k_via) - 1

        major_scales = np.empty((n_via+1,))
        minor_scales = np.empty((n_via * n_split,))
        k = np.empty((n_via * n_split, self.n_dim))

        count = 0
        length_part = 0.0
        major_scales[0] = 0.0

        for i in range(n_via):
            direction = k_via[i+1] - k_via[i]
            length = np.linalg.norm(direction)

            x = np.linspace(0.0, 1.0, n_split)
            
            for j in range(n_split):
                k[count] = k_via[i] + x[j] * direction
                minor_scales[count] = x[j] * length + length_part
                count += 1
            
            length_part += length
            major_scales[i+1] = length_part

        k_path = PathValues(major_scales, minor_scales, k)

        return k_path
    
    def get_reciprocal_vectors(self, n_g: int) -> np.ndarray:
        """Generate the reciprocal lattice vectors used to define the Brillouin zone boundaries.
        
        Args:
            n_g: Number of reciprocal lattice vectors

        Returns:
            np.ndarray: A numpy array of reciprocal lattice vectors.
        """
        if self.n_dim == 3:
            n_cut = np.ceil(np.cbrt(n_g))
        elif self.n_dim == 2:
            n_cut = np.ceil(np.sqrt(n_g))
        elif self.n_dim == 1:
            n_cut = np.ceil(np.abs(n_g))
        else:
            raise ValueError("n_cut = 1, 2, 3")

        n_array = np.arange(-n_cut, n_cut + 1)

        grids = np.meshgrid(*([n_array] * self.n_dim), indexing='ij')
        
        grid_points = np.stack(grids, axis=-1).reshape(-1, self.n_dim)
        
        g = grid_points @ self.basis
        g_norm = np.linalg.norm(g, axis=-1).round(decimals=5)
        g_norm_unique = np.unique(g_norm)

        g_list = []

        for g_ref in g_norm_unique:
            count = 0
            degeneracy = 0
            for g_compare in g_norm:
                if g_compare == g_ref:
                    g_list.append(g[count])
                    degeneracy += 1
                count += 1

        return np.array(g_list[0:n_g])
    
    def calculate_special_k(k_name: str) -> None:
        pass

@dataclass
class ReciprocalCell3D(ReciprocalCell, Cell3D):
    """Class for the 3D reciprocal cell for a crystal based on the lattice constants of the primitive cell.
    
    Attributes:
        lattice_constant (LatticeConstant3D): Lattice constants.
        basis (np.ndarray): A numpy array of basis.
        volume (float): The volume.
    """
    lattice_constant: LatticeConstant3D
    
    def __post_init__(self):
        """Initializes and builds the basis for the reciprocal cell."""
        Cell3D.__init__(self)
        self.basis = self.build()
        self.volume = self.calculate_volume()
    
    def build(self) -> np.ndarray:
        """Constructs the basis matrix for the reciprocal cell from the primitive cell.

        Returns:
            basis (np.ndarray): The basis matrix of the reciprocal cell.
        """
        primitive_cell = PrimitiveCell3D(self.lattice_constant)

        basis = np.zeros((3,3))
        
        primitive_vector = primitive_cell.basis
        for i in range(3):
            j = (i+1) % 3
            k = (i+2) % 3
            basis[i] = np.cross(primitive_vector[j], primitive_vector[k])

        basis *= 2.0 * np.pi / primitive_cell.volume
        
        return basis

    def get_monkhorst_pack_grid(self, n_x: int, n_y: int, n_z: int) -> np.ndarray:
        """Get 3D Monkhorst and Pack grid

        Args:
            n_x (int): Number of x-direction points
            n_y (int): Number of y-direction points
            n_z (int): Number of z-direction points

        Returns:
            aligned_k (np.ndarray): A numpy array of k points
        """
        x = (2 * np.arange(1, n_x + 1) - n_x - 1) / (2 * n_x)
        y = (2 * np.arange(1, n_y + 1) - n_y - 1) / (2 * n_y)
        z = (2 * np.arange(1, n_z + 1) - n_z - 1) / (2 * n_z)

        bx = np.broadcast_to(x[:, np.newaxis, np.newaxis], (n_x, n_y, n_z))
        by = np.broadcast_to(y[np.newaxis, :, np.newaxis], (n_x, n_y, n_z))
        bz = np.broadcast_to(z[np.newaxis, np.newaxis, :], (n_x, n_y, n_z))

        aligned_k = np.stack([bx, by, bz], axis=-1).reshape(-1, 3) @ self.basis

        return aligned_k

    def calculate_special_k(self, k_name: str) -> np.ndarray:
        """Retrieves the coordinates of special k-points based on the crystal structure.

        Args:
            k_name (str): The name of the special point.

        Returns:
            np.ndarray: The coordinates of the special point.

        Raises:
            ValueError: If an invalid crystal structure name is provided.
        """
        if self.lattice_constant.crystal_structure == 'bcc':
            special_point = SpecialPoints3D.BCC[k_name]
        elif self.lattice_constant.crystal_structure == 'fcc':
            special_point = SpecialPoints3D.FCC[k_name]
        elif self.lattice_constant.crystal_structure == 'sc':
            special_point = SpecialPoints3D.SC[k_name]
        else:
            raise ValueError("Invalid name specified.")
        
        return np.array(special_point) @ self.basis

@dataclass
class ReciprocalCell2D(ReciprocalCell, Cell2D):
    """Class for the 2D reciprocal cell for a crystal based on the lattice constants of the primitive cell.
    
    Attributes:
        lattice_constant (LatticeConstant2D): Lattice constants.
        basis (np.ndarray): A numpy array of basis.
        volume (float): The volume.
    """
    lattice_constant: LatticeConstant2D
    
    def __post_init__(self):
        """Initializes and builds the basis for the reciprocal cell."""
        Cell2D.__init__(self)
        self.basis = self.build()
        self.volume = self.calculate_volume()
    
    def build(self) -> np.ndarray:
        """Constructs the basis matrix for the reciprocal cell from the primitive cell.

        Returns:
            basis (np.ndarray): The basis matrix of the reciprocal cell.
        """
        primitive_cell = PrimitiveCell2D(self.lattice_constant)

        basis = np.zeros((2,2))
        
        primitive_vector = primitive_cell.basis

        quarter_rotation_matrix = np.array([
            [0.0, -1.0],
            [1.0, 0.0]
        ])
        
        for i in range(2):
            j = (i+1) % 2
            rotated_primitive_vector = quarter_rotation_matrix @ primitive_vector[j]
            basis[i] = rotated_primitive_vector / np.dot(primitive_vector[i], rotated_primitive_vector) * 2.0 * np.pi
        
        return basis

    def get_monkhorst_pack_grid(self, n_x: int, n_y: int) -> np.ndarray:
        """Get 2D Monkhorst and Pack grid

        Args:
            n_x (int): Number of x-direction points
            n_y (int): Number of y-direction points

        Returns:
            aligned_k (np.ndarray): A numpy array of k points
        """

        x = (2 * np.arange(1, n_x + 1) - n_x - 1) / (2 * n_x)
        y = (2 * np.arange(1, n_y + 1) - n_y - 1) / (2 * n_y)

        bx = np.broadcast_to(x[:, np.newaxis], (n_x, n_y))
        by = np.broadcast_to(y[np.newaxis, :], (n_x, n_y))

        aligned_k = np.stack([bx, by], axis=-1).reshape(-1, 2) @ self.basis

        return aligned_k

    def calculate_special_k(self, k_name: str) -> np.ndarray:
        """Retrieves the coordinates of special k-points based on the crystal structure.

        Args:
            k_name (str): The name of the special point.

        Returns:
            np.ndarray: The coordinates of the special point.

        Raises:
            ValueError: If an invalid crystal structure name is provided.
        """
        if self.lattice_constant.crystal_structure == 'square':
            special_point = SpecialPoints2D.Square[k_name]
        elif self.lattice_constant.crystal_structure == 'hexagonal':
            special_point = SpecialPoints2D.Hexagonal[k_name]
        else:
            raise ValueError("Invalid name specified.")
        
        return np.array(special_point) @ self.basis

@dataclass
class ReciprocalCell1D(ReciprocalCell, Cell1D):
    """Class for the 1D reciprocal cell for a crystal based on the lattice constants of the primitive cell.
    
    Attributes:
        lattice_constant (LatticeConstant1D): Lattice constants.
        basis (np.ndarray): A numpy array of basis.
        volume (float): The volume.
    """

    lattice_constant: LatticeConstant1D
    
    def __post_init__(self):
        """Initializes and builds the basis for the reciprocal cell."""
        Cell1D.__init__(self)
        self.basis = self.build()
        self.volume = self.calculate_volume()
    
    def build(self) -> np.ndarray:
        """Constructs the basis matrix for the reciprocal cell from the primitive cell.

        Returns:
            np.ndarray: The basis matrix of the reciprocal cell.
        """
        primitive_cell = PrimitiveCell1D(self.lattice_constant)
        
        primitive_vector = primitive_cell.basis

        basis = 2.0 * np.pi / primitive_vector
        
        return basis

    def get_monkhorst_pack_grid(self, n_x: int) -> np.ndarray:
        """Get 1D Monkhorst and Pack grid

        Args:
            n_x (int): Number of x-direction points

        Returns:
            aligned_k (np.ndarray): A numpy array of k points
        """

        x = (2 * np.arange(1, n_x + 1) - n_x - 1) / (2 * n_x)

        aligned_k = np.stack([x], axis=-1).reshape(-1, 1) @ self.basis
        
        return aligned_k

    def calculate_special_k(self, k_name: str) -> np.ndarray:
        """Retrieves the coordinates of special k-points based on the crystal structure.

        Args:
            k_name (str): The name of the special point.

        Returns:
            np.ndarray: The coordinates of the special point.

        Raises:
            ValueError: If an invalid crystal structure name is provided.
        """
        special_point = SpecialPoints1D.Line[k_name]
        
        return special_point * self.basis


import numpy as np

from elphem.common.function import safe_divide
from elphem.common.distribution import fermi_distribution, bose_distribution
from elphem.electron.electron import Electron
from elphem.phonon.phonon import Phonon

class GreenFunction:
    """Class for Green functions
        Attributes:
            eta (float): Convergence factor for calculating real parts
            gaussian_coefficient_a (float): Coefficient for calculating imaginary parts (a)
            gaussian_coefficient_b (float): Coefficient for calculating imaginary parts (b)
            poles (np.ndarray): A numpy array of poles
            weights (np.ndarray): A numpy array of weights for each pole
    """
    def __init__(self, electron: Electron, phonon: Phonon, temperature: float, sigma: float, eta: float):
        # set attributes
        self.eta = eta
        self.gaussian_coefficient_a = 2.0 * sigma ** 2
        self.gaussian_coefficient_b = np.sqrt(2.0 * np.pi) * sigma
        self.poles = np.array([
            electron.eigenenergies + phonon.eigenenergies,
            electron.eigenenergies - phonon.eigenenergies
            ])
        
        # prepare occupations
        electron_occupations = fermi_distribution(temperature, electron.eigenenergies)
        phonon_occupations = bose_distribution(temperature, phonon.eigenenergies)

        # set weights
        self.weights = np.array([
            1.0 - electron_occupations + phonon_occupations,
            electron_occupations + phonon_occupations
            ])
    
    def calculate(self, omega: float) -> np.ndarray:
        """Calculate Green function

        Args:
            omega (float): A frequency

        Returns:
            np.ndarray: A numpy array of Green function (complex)
        """
        omega_minus_poles = omega - self.poles
        
        return self.calculate_real_part(omega_minus_poles) + 1.0j * self.calculate_imag_part(omega_minus_poles)
    
    def calculate_real_part(self, omega_minus_poles: np.ndarray) -> np.ndarray:
        """Calculate real part of Green function

        Args:
            omega_minus_poles (np.ndarray): A numpy array of frequencies

        Returns:
            np.ndarray: A numpy array of real part of Green function
        """
        real_part = np.nansum(safe_divide(self.weights, omega_minus_poles + 1.0j * self.eta).real, axis=0)
        
        return real_part
    
    def calculate_imag_part(self, omega_minus_poles: np.ndarray) -> np.ndarray:
        """Calculate imaginary part of Green function

        Args:
            omega_minus_poles (np.ndarray): A numpy array of frequencies

        Returns:
            np.ndarray: A numpy array of imaginary part of Green function
        """
        imag_part = - np.nansum(self.weights * np.exp(- omega_minus_poles ** 2 / self.gaussian_coefficient_a), axis=0) / self.gaussian_coefficient_b * np.pi
        
        return imag_part


import numpy as np

from elphem.electron.electron import Electron
from elphem.phonon.phonon import Phonon
from elphem.elph.green_function import GreenFunction

from elphem.common.stdout import ProgressBar
from elphem.common.function import safe_divide


class ElectronPhonon:
    """Calculate the electron-phonon interactions.

    Attributes:
        temperature (float): Temperature of the system in Kelvin
        sigma (float): Smearing parameter for the Gaussian distribution, defaults to 0.001 Hartree
        eta (float): Small positive constant to ensure numerical stability, defaults to 0.0005 Hartree
        effective_potential (float): Effective potential used in electron-phonon coupling calculation, defaults to 0.0625
        n_bands (int): Number of bands for calculating self energies
        electron (Electron): Free electron for initial and final states
        electron_inter (Electron): Free electron for intermediate states
        phonon (Phonon): Debye Phonon
        green_function (GreenFunction): Green function
        couplings (np.ndarray): electron-phonon coupling constants squared
    """
    def __init__(self, electron: Electron, phonon: Phonon, temperature: float, n_bands: int,
                sigma: float = 0.001, eta: float = 0.0005, coupling_type: str = "bardeen"):
        self.n_dim = electron.lattice.n_dim
        self.temperature = temperature
        if n_bands > electron.n_bands:
            self.n_bands = electron.n_bands
        else:
            self.n_bands = n_bands

        self.eigenenergies = electron.eigenenergies[0:self.n_bands, :]
        
        g1, g2, k, q = self.create_ggkq_grid(electron, phonon)

        # set electrons and a phonon
        self.electron = electron.clone_with_gk_grid(g1, k)
        self.electron_inter = electron.clone_with_gk_grid(g2, k + q)
        self.phonon = phonon.clone_with_q_grid(q)
        
        # set Green function
        self.green_function = GreenFunction(self.electron_inter, self.phonon, self.temperature, sigma, eta)

        # set electron-phonon coupling constants squared
        self.coupling2 = np.abs(self.calculate_couplings(coupling_type)) ** 2

    def create_ggkq_grid(self, electron: Electron, phonon: Phonon) -> tuple:
        """Create (G_!, G_2, k, q) combined grids

        Args:
            electron (Electron): Free electron
            phonon (Phonon): Debye phonon

        Returns:
            tuple: (G_1, G_2, k, q) combined grids
        """
        shape = (self.n_bands, electron.n_bands, electron.n_k, phonon.n_q, self.n_dim)
        
        g1 = np.broadcast_to(electron.g[:self.n_bands, np.newaxis, np.newaxis, np.newaxis, :], shape)
        g2 = np.broadcast_to(electron.g[np.newaxis, :, np.newaxis, np.newaxis, :], shape)
        k = np.broadcast_to(electron.k[np.newaxis, np.newaxis, :, np.newaxis, :], shape)
        q = np.broadcast_to(phonon.q[np.newaxis, np.newaxis, np.newaxis, :, :], shape)
        
        return g1, g2, k, q

    def calculate_couplings(self, coupling_type: str) -> np.ndarray:
        if coupling_type == "bloch":
            return self.calculate_couplings_bloch()
        elif coupling_type == "nordheim":
            return self.calculate_couplings_nordheim()
        elif coupling_type == "bardeen":
            return self.calculate_couplings_bardeen()
        else:
            raise ValueError("coupling_type is invalid.")

    def calculate_couplings_bloch(self) -> np.ndarray:
        """Calculate the Bloch (1929) electron-phonon coupling constants.

        Returns:
            np.ndarray: The lowest-order electron-phonon coupling constants
        """
        potential = 1.0 / 16.0
        
        return -1.0j * potential * np.nansum((self.phonon.q + self.electron_inter.g - self.electron.g) * self.phonon.eigenvectors, axis=-1) * self.phonon.zero_point_lengths

    def calculate_couplings_nordheim(self) -> np.ndarray:
        """Calculate the Nordheim (1931) electron-phonon coupling constants.

        Returns:
            np.ndarray: The lowest-order electron-phonon coupling constants
        """
        potential = 4.0 / self.electron.lattice.primitive.volume * self.electron.n_electrons * np.pi / ( np.nansum(self.phonon.q + self.electron_inter.g - self.electron.g, axis=-1) ** 2)
        
        return -1.0j * potential * np.nansum((self.phonon.q + self.electron_inter.g - self.electron.g) * self.phonon.eigenvectors, axis=-1) * self.phonon.zero_point_lengths

    def calculate_couplings_bardeen(self) -> np.ndarray:
        """Calculate the Bardeen (1937) electron-phonon coupling constants.

        Returns:
            np.ndarray: The lowest-order electron-phonon coupling constants
        """
        wave_number = np.nansum(self.phonon.q + self.electron_inter.g - self.electron.g, axis=-1)

        numerator = 4.0 * self.electron.n_electrons * np.pi
        denominator = wave_number ** 2 + self.electron.thomas_fermi_wave_number ** 2 * self.calculate_lindhard_function(wave_number / (2.0 * self.electron.fermi_wave_number))
        
        return -1.0j / self.electron.lattice.primitive.volume * numerator / denominator * np.nansum((self.phonon.q + self.electron_inter.g - self.electron.g) * self.phonon.eigenvectors, axis=-1) * self.phonon.zero_point_lengths
    
    @staticmethod
    def calculate_lindhard_function(x: np.ndarray) -> np.ndarray:
        return 0.5 + (1.0 - x ** 2) / (4.0 * x) * np.log(np.abs(1.0 + x) / np.abs(1.0 - x))

    def calculate_self_energies(self, omega: float) -> np.ndarray:
        """Calculate Fan self-energies.

        Args:
            omega (float): a frequency.

        Returns:
            np.ndarray: A numpy array of Fan self-energies.
        """
        
        return np.nansum(self.coupling2 * self.green_function.calculate(omega), axis=(1, 3)) / self.phonon.n_q

    def calculate_electron_phonon_renormalization(self) -> np.ndarray:
        """Calculate electron-phonon renormalizations of electron eigenenergies (EPR).

        Returns:
            np.ndarray: A numpy array of EPR
        """
        # prepare an array for EPR
        epr = np.empty(self.eigenenergies.shape)
        
        # calculate EPR
        count = 0
        progress_bar = ProgressBar('Electron Phonon Renormalization', self.n_bands * self.electron.n_k)
        for i in range(self.n_bands):
            for j in range(self.electron.n_k):
                self_energies = self.calculate_self_energies(self.eigenenergies[i, j])
                epr[i, j] = self_energies[i, j].real
                count += 1
                progress_bar.print(count)

        return epr

    def calculate_spectrum(self, omega: float) -> np.ndarray:
        """Calculate the spectral function for a frequency.

        Args:
            omega (float): a frequency

        Returns:
            np.ndarray: A numpy array of the spectral function at omega Hartree
        """
        # calculate self energies at omega
        self_energies = self.calculate_self_energies(omega)
        
        # calculate numerator and denominator separately
        numerator = - self_energies.imag / np.pi
        denominator = (omega - self.eigenenergies - self_energies.real) ** 2 + self_energies.imag ** 2
        
        # sum over bands
        return np.nansum(safe_divide(numerator, denominator), axis=0)

    def calculate_self_energies_over_range(self, omega_array: np.ndarray | list[float]) -> np.ndarray:
        """Calculate self energies over a given array of frequencies.

        Args:
            omega_array (np.ndarray | list[float]): a numpy array or list of frequencies

        Returns:
            np.ndarray: a numpy array of self energies
        """
        # prepare the length of frequency-array and an array for self energies
        n_omega = len(omega_array)
        self_energies = np.empty(self.eigenenergies.shape + (n_omega,), dtype='complex')
        
        # calculate self energies
        count = 0
        progress_bar = ProgressBar('Self Energy', n_omega)
        for omega in omega_array:
            self_energies[..., count] = self.calculate_self_energies(omega)
            
            count += 1
            progress_bar.print(count)

        return self_energies
        
    def calculate_spectrum_over_range(self, omega_array: np.ndarray | list[float], normalize: bool = False) -> np.ndarray:
        """Calculate spectral function over a given array of frequencies.

        Args:
            omega_array (np.ndarray | list[float]): A numpy array or list of frequencies
            normalize (bool, optional): Flag for normalizing spectral functions. Defaults to False.

        Returns:
            np.ndarray: A numpy array of spectral functions
        """
        # prepare
        n_omega = len(omega_array)
        spectrum = np.empty((self.electron.n_k, n_omega))
        
        # calculate spectral functions
        progress_bar = ProgressBar('Spectrum', n_omega)
        for i in range(n_omega):
            spectrum[..., i] = self.calculate_spectrum(omega_array[i])

            progress_bar.print(i)
        
        # normalization
        if normalize:
            spectrum_sum = np.nansum(spectrum, axis=1)
            for i in range(n_omega):
                spectrum[..., i] /= spectrum_sum

        return spectrum

    def calculate_coupling_strengths(self, delta_omega: float = 0.000001) -> np.ndarray:
        """Calculate electron-phonon coupling strengths

        Args:
            delta_omega (float, optional): A small value used for numerical differentiation. Defaults to 0.000001 Hartree.

        Returns:
            np.ndarray: A numpy array of coupling strengths
        """
        # prepare an array
        coupling_strengths = np.empty(self.eigenenergies.shape)
        
        # calculate coupling strengths
        count = 0
        progress_bar = ProgressBar('Coupling Strength', self.n_bands * self.electron.n_k)
        for i in range(self.n_bands):
            for j in range(self.electron.n_k):
                # numerical differentiation
                self_energies_plus = self.calculate_self_energies(self.eigenenergies[i,j] + delta_omega)
                self_energies_minus = self.calculate_self_energies(self.eigenenergies[i,j] - delta_omega)
                coupling_strengths[i,j] = - (self_energies_plus[i,j].real - self_energies_minus[i,j].real) / (2.0 * delta_omega)
                
                count += 1
                progress_bar.print(count)
        
        return coupling_strengths


import math
import scipy.constants as constants

class Prefix:
    """ Metric prefixes. """
    EXA = 1.0e+18
    PETA = 1.0e+15
    TERA = 1.0e+12
    GIGA = 1.0e+9
    MEGA = 1.0e+6
    KILO = 1.0e+3
    HECTO = 1.0e+2
    DECA = 1.0e+1
    DECI = 1.0e-1
    CENTI = 1.0e-2
    MILLI = 1.0e-3
    MICRO = 1.0e-6
    NANO = 1.0e-9
    PICO = 1.0e-12
    FEMTO = 1.0e-15
    ATTO = 1.0e-18

class Byte:
    """ Digital information units based on powers of 2. """
    KILO = 2 ** 10
    MEGA = 2 ** 20
    GIGA = 2 ** 30
    TERA = 2 ** 40
    PETA = 2 ** 50
    EXA = 2 ** 60

    @staticmethod
    def get_str(size: int) -> str:
        """Returns a human-readable string representation of digital information sizes, scaled to appropriate unit.

        Args:
            size (int): Size in bytes.

        Returns:
            str: Formatted string representing the size in appropriate unit (e.g., KB, MB).

        Raises:
            ValueError: If size is negative or too large to be represented.
        """
        if size < 0:
            raise ValueError("Size should be a non-negative integer.")
        
        units = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB")
        i = math.floor(math.log(size, 1024)) if size > 0 else 0
        
        if i >= len(units):
            raise ValueError("Size too large to represent in known units.")
        
        size = round(size / 1024 ** i, 2)
        return f"{size} {units[i]}"

class AtomicUnits:
    """ Utility class for atomic unit conversions. """
    @staticmethod
    def convert(unit: float, value: float) -> dict:
        """Converts a value to and from a base unit to facilitate comparisons and calculations.

        Args:
            unit (float): The unit to which the conversion is based.
            value (float): The value to be converted.

        Returns:
            dict: A dictionary containing converted values.
        """
        return {
            "->": value / unit,
            "<-": unit / value
        }

class Length:
    """ Atomic units of length. """
    UNIT = 5.29177210903e-11
    SI = AtomicUnits.convert(UNIT, 1.0)
    ANGSTROM = AtomicUnits.convert(UNIT, 1.0e-10)

class Energy:
    """ Atomic units of energy. """
    UNIT = 4.3597447222071e-18
    SI = AtomicUnits.convert(UNIT, 1.0)
    EV = AtomicUnits.convert(UNIT, constants.eV)
    KELVIN = AtomicUnits.convert(UNIT, constants.k)

class Mass:
    """ Atomic units of mass. """
    UNIT = 9.1093837015e-31
    SI = AtomicUnits.convert(UNIT, 1.0)
    DALTON = AtomicUnits.convert(UNIT, 1.66053906660e-27)
    PROTON = AtomicUnits.convert(UNIT, 1.672621898e-27)

class Time:
    """ Atomic units of time. """
    UNIT = 2.4188843265857e-17
    SI = AtomicUnits.convert(UNIT, 1.0)

class Velocity:
    """ Atomic units of velocity. """
    UNIT = 2.18769126364e+6
    SI = AtomicUnits.convert(UNIT, 1.0)

class ElectricPotential:
    """ Atomic units of electric potential. """
    UNIT = 27.211386245988
    SI = AtomicUnits.convert(UNIT, 1.0)


import numpy as np

class Cell3D:
    """Class for 3D cell
    Attributes:
        n_dim (int): Number of dimension
        basis (np.ndarray): A numpy array of basis
    """
    def __init__(self):
        self.n_dim = 3
        self.basis = None
    
    def calculate_volume(self) -> float:
        """Calculates the volume of the cell.

        Returns:
            float: The volume.
        """
        return np.dot(self.basis[0], np.cross(self.basis[1], self.basis[2]))

class Cell2D:
    """Class for 2D cell
    Attributes:
        n_dim (int): Number of dimension
        basis (np.ndarray): A numpy array of basis
    """
    def __init__(self):
        self.n_dim = 2
        self.basis = None
    
    def calculate_volume(self) -> float:
        """Calculates the volume of the cell.

        Returns:
            float: The volume.
        """
        return np.linalg.norm(np.cross(self.basis[0], self.basis[1]))

class Cell1D:
    """Class for 1D cell
    Attributes:
        n_dim (int): Number of dimension
        basis (np.ndarray): A numpy array of basis
    """

    def __init__(self):
        self.n_dim = 1
        self.basis = None
    
    def calculate_volume(self) -> float:
        """Calculates the volume of the cell.

        Returns:
            float: The volume.
        """
        return np.linalg.norm(self.basis)


import numpy as np

def safe_divide(a: np.ndarray | float | int, b: np.ndarray | float | int, default=float('nan')) -> np.ndarray:
    """
    Safely divides two numbers, arrays, or a combination thereof, with optional handling of division by zero.

    Args:
        a (np.ndarray | complex | float | int): Numerator, can be a number or an array.
        b (np.ndarray | complex | float | int): Denominator, can be a number or an array.
        default (complex | float | int, optional): Value to use when division by zero occurs. Default is NaN.

    Returns:
        np.ndarray: Result of the division (a / b), with division by zero handled gracefully.
    """

    dtype = np.result_type(a, b, default)
    
    a_array = np.full_like(b, a, dtype=dtype) if np.isscalar(a) else a.astype(dtype)
    
    out = np.full_like(b, fill_value=default, dtype=dtype)
    
    with np.errstate(divide='ignore', invalid='ignore'):
        result = np.divide(a_array, b, out=out, where=b != 0)
        
    return result


from unittest import TestCase
from elphem.common.unit import *

class TestUnit(TestCase):
    def test_byte(self):
        size = Byte.get_str(2**30)
        self.assertEqual(size, "1.0 GB")
    
    def test_length(self):
        self.assertEqual(Length.SI["<-"], Length.UNIT)
    
    def test_energy(self):
        self.assertEqual(Energy.SI["<-"], Energy.UNIT)
        self.assertEqual(Energy.EV["->"] * Energy.SI["<-"], 1.602176634e-19)
        self.assertEqual(Energy.KELVIN["->"] * Energy.KELVIN["<-"], 1.0)


import numpy as np
from dataclasses import dataclass

from elphem.lattice.rotation import LatticeRotation3D, LatticeRotation2D
from elphem.lattice.cell import Cell1D, Cell2D, Cell3D
from elphem.lattice.lattice_constant import LatticeConstant3D, LatticeConstant2D, LatticeConstant1D

@dataclass
class PrimitiveCell3D(Cell3D):
    """Class for the 3D primitive cell based on the lattice constants.

    Attributes:
        lattice_constant (LatticeConstant3D): lattice constants
        basis (np.ndarray): A numpy array of basis
        volume (float): The volume
    """
    lattice_constant: LatticeConstant3D
    
    def __post_init__(self):
        """Initializes and builds the basis for the primitive cell."""
        self.basis = self.build()
        self.volume = self.calculate_volume()
    
    def build(self) -> np.ndarray:
        """Constructs the basis matrix for the primitive cell from lattice constants and angles.

        Returns:
            np.ndarray: The basis matrix of the primitive cell.
        """
        length = self.lattice_constant.length
        angle = self.lattice_constant.angle

        basis = np.zeros((3,3))

        basis[0][0] = length[0]
        basis[1][0] = length[1] * np.cos(angle[2])
        basis[1][1] = length[1] * np.sin(angle[2])
        basis[2][0] = length[2] * np.cos(angle[1])
        basis[2][1] = length[2] * (np.cos(angle[0]) - np.cos(angle[1]) * np.cos(angle[2])) / np.sin(angle[2])
        basis[2][2] = np.sqrt(length[2] ** 2 - np.sum(basis[2]**2))

        basis = self.optimize(basis)
        
        return basis

    @staticmethod
    def optimize(basis: np.ndarray) -> np.ndarray:
        """Optimizes the cell basis using the LatticeRotation utility.

        Args:
            basis (np.ndarray): The basis matrix to optimize.

        Returns:
            np.ndarray: The optimized basis matrix.
        """
        axis = np.array([1.0] * 3)
        
        return LatticeRotation3D.calculate_optimized_basis(basis, axis)

@dataclass
class PrimitiveCell2D(Cell2D):
    """Class for the 2D primitive cell based on the lattice constants.

    Attributes:
        lattice_constant (LatticeConstant3D): lattice constants
        basis (np.ndarray): A numpy array of basis
        volume (float): The volume
    """

    lattice_constant: LatticeConstant2D

    def __post_init__(self):
        """Initializes and builds the basis for the primitive cell."""
        self.basis = self.build()
        self.volume = self.calculate_volume()
    
    def build(self) -> np.ndarray:
        """Constructs the basis matrix for the primitive cell from lattice constants and angles.

        Returns:
            np.ndarray: The basis matrix of the primitive cell.
        """
        length = self.lattice_constant.length
        angle = self.lattice_constant.angle

        basis = np.zeros((2,2))

        basis[0][0] = length[0]
        basis[1][0] = length[1] * np.cos(angle)
        basis[1][1] = length[1] * np.sin(angle)

        basis = self.optimize(basis)
        
        return basis

    @staticmethod
    def optimize(basis: np.ndarray) -> np.ndarray:
        """Optimizes the cell basis using the LatticeRotation utility.

        Args:
            basis (np.ndarray): The basis matrix to optimize.

        Returns:
            np.ndarray: The optimized basis matrix.
        """
        axis = np.array([1.0, 1.0])
        
        return LatticeRotation2D.calculate_optimized_basis(basis, axis)

@dataclass
class PrimitiveCell1D(Cell1D):
    """Class for the 1D primitive cell based on the lattice constants.

    Attributes:
        lattice_constant (LatticeConstant3D): lattice constants
        basis (np.ndarray): A numpy array of basis
        volume (float): The volume
    """

    lattice_constant: LatticeConstant1D

    def __post_init__(self):
        """Initializes and builds the basis for the primitive cell."""
        self.basis = self.build()
        self.volume = self.calculate_volume()
    
    def build(self) -> np.ndarray:
        """Constructs the basis matrix for the primitive cell from lattice constants and angles.

        Returns:
            np.ndarray: The basis matrix of the primitive cell.
        """
        basis = np.array([[self.lattice_constant.length]])
        
        return basis


import numpy as np
from unittest import TestCase

from elphem.common.unit import Length
from elphem.lattice.lattice import Lattice3D
from elphem.electron.electron import Electron
from elphem.phonon.phonon import Phonon
from elphem.elph.electron_phonon import ElectronPhonon

class TestUnit(TestCase):
    def setUp(self) -> None:
        n_electrons = 1
        n_bands_electrons = 10
        n_bands_elph = 2
        a = 2.98 * Length.ANGSTROM["->"]
        n_k = np.full(3,4)
        n_q = np.full(3,4)
        
        lattice = Lattice3D('bcc', 'Li', a)

        debye_temperature = 344.0
        temperature = 0.3 * debye_temperature

        self.electron = Electron.create_from_n(lattice, n_electrons, n_bands_electrons, n_k)
        self.phonon = Phonon.create_from_n(lattice, debye_temperature, n_q)
        self.electron_phonon = ElectronPhonon(self.electron, self.phonon, temperature, n_bands_elph)
    
    def test_coupling(self):
        coupling2 = self.electron_phonon.coupling2
        
        self.assertEqual(coupling2.shape, (self.electron_phonon.n_bands, self.electron.n_bands, self.electron.n_k, self.phonon.n_q))
    
    def test_self_energies(self):
        self_energies = self.electron_phonon.calculate_self_energies(0.0)
        
        self.assertEqual(self_energies.shape, (self.electron_phonon.n_bands, self.electron.n_k))
    
    def test_spectrum(self):
        spectrum = self.electron_phonon.calculate_spectrum(0.0)
        
        self.assertEqual(spectrum.shape, (self.electron.n_k))
    
    def test_self_energies_over_range(self):
        omega_array = np.linspace(-1.0, 1.0, 50)
        self_energies = self.electron_phonon.calculate_self_energies_over_range(omega_array)
        
        self.assertEqual(self_energies.shape, (self.electron_phonon.n_bands, self.electron.n_k, len(omega_array)))
    
    def test_spectrum_over_range(self):
        omega_array = np.linspace(-1.0, 1.0, 50)
        self_energies = self.electron_phonon.calculate_spectrum_over_range(omega_array)
        
        self.assertEqual(self_energies.shape, (self.electron.n_k, len(omega_array)))


from unittest import TestCase
from elphem.common.atomic_weight import *

class TestUnit(TestCase):
    def test_atomic_weight(self):
        self.assertEqual(AtomicWeight.table["C"], 12.011)

        atom_name_list = ["Ga", "As"]
        weight_list = AtomicWeight.get_from_list(atom_name_list)
        ref_list = [AtomicWeight.table[s] for s in atom_name_list]
        self.assertEqual(weight_list, ref_list)


from types import MappingProxyType

class SpecialPoints3D:
    """Defines immutable sets of special points in first Brillouin zone for different types of 3D crystal lattices.

    Attributes:
        Gamma (MappingProxyType): Immutable set containing the Gamma point.
        SC (MappingProxyType): Immutable set of special points for the simple cubic lattice.
        FCC (MappingProxyType): Immutable set of special points for the face-centered cubic lattice.
        BCC (MappingProxyType): Immutable set of special points for the body-centered cubic lattice.
        Hexagonal (MappingProxyType): Immutable set of special points for the hexagonal lattice.
    """
    
    # Gamma Point
    _Gamma = {
        "G": (0, 0, 0)
    }
    Gamma = MappingProxyType(_Gamma)

    # Simple Cubic
    _SC = {**_Gamma}
    _SC.update({
        "R": (1/2, 1/2, 1/2),
        "X": (0, 1/2, 0),
        "M": (1/2, 1/2, 0)
    })
    SC = MappingProxyType(_SC)

    # Face-Centered Cubic
    _FCC = {**_Gamma}
    _FCC.update({
        "X": (0, 1/2, 1/2),
        "L": (1/2, 1/2, 1/2),
        "W": (1/4, 3/4, 1/2),
        "U": (1/4, 5/8, 5/8),
        "K": (3/8, 3/4, 3/8)
    })
    FCC = MappingProxyType(_FCC)

    # Body-Centered Cubic
    _BCC = {**_Gamma}
    _BCC.update({
        "H": (-1/2, 1/2, 1/2),
        "P": (1/4, 1/4, 1/4),
        "N": (0, 1/2, 0)
    })
    BCC = MappingProxyType(_BCC)

    # Hexagonal
    _Hexagonal = {**_Gamma}
    _Hexagonal.update({
        "A": (0, 0, 1/2),
        "K": (2/3, 1/3, 0),
        "H": (2/3, 1/3, 1/2),
        "M": (1/2, 0, 0),
        "L": (1/2, 0, 1/2)
    })
    Hexagonal = MappingProxyType(_Hexagonal)

class SpecialPoints2D:
    """Defines immutable sets of special points in first Brillouin zone for different types of 2D crystal lattices.

    Attributes:
        Gamma (MappingProxyType): Immutable set containing the Gamma point.
        Square (MappingProxyType): Immutable set of special points for the square lattice.
        Hexagonal (MappingProxyType): Immutable set of special points for the hexagonal lattice.
    """

    # Gamma Point
    _Gamma = {
        "G": (0, 0)
    }
    Gamma = MappingProxyType(_Gamma)

    # Square
    _Square = {**_Gamma}
    _Square.update({
        "X": (1/2, 0),
        "M": (1/2, 1/2)
    })
    Square = MappingProxyType(_Square)

    # Hexagonal
    _Hexagonal = {**_Gamma}
    _Hexagonal.update({
        "M": (1/2, 0),
        "K": (-1/3, 2/3)
    })
    Hexagonal = MappingProxyType(_Hexagonal)

class SpecialPoints1D:
    """Defines immutable sets of special points in first Brillouin zone for different types of 1D crystal lattices.

    Attributes:
        Gamma (MappingProxyType): Immutable set containing the Gamma point.
        Line (MappingProxyType): Immutable set of special points for the lattice.
    """

    # Gamma Point
    _Gamma = {
        "G": 0
    }
    Gamma = MappingProxyType(_Gamma)
    
    # Line
    _Line = {**_Gamma}
    _Line.update({
        "X": 1/2
    })
    Line = MappingProxyType(_Line)


import numpy as np

from elphem.common.unit import Energy
from elphem.common.function import safe_divide
from elphem.lattice.lattice import Lattice
from elphem.lattice.path import PathValues

class Phonon:
    """ Phonon with Debye model.

    Attributes:
        lattice (Lattice): A given empty lattice.
        debye_temperature (float): The Debye temperature for phonons.
        q (np.ndarray): Vectors in a given reciprocal space.
        n_q (int): Number of q points.
        eigenenergies (np.ndarray): Eigenenergies of phonon in Debye model.
        eigenvectors (np.ndarray): Eigenvectors of phonon in Debye model.
        zero_point_lengths (np.ndarray): Zero point lengths in Debye model.
        speed_of_sound (float): Speed of sound
    """

    def __init__(self, lattice: Lattice, debye_temperature: float) -> None:
        self.lattice = lattice
        self.debye_temperature = debye_temperature
        
        self.q = None
        self.n_q = None
        self.eigenenergies = None
        self.eigenvectors = None
        self.zero_point_lengths = None

        self.speed_of_sound = self.calculate_speed_of_sound()
    
    @classmethod
    def create_from_q(cls, lattice: Lattice, debye_temperature: float, q_array: np.ndarray) -> 'Phonon':
        """Create Debye phonon from q vectors

        Args:
            lattice (Lattice): Lattice on which the Debye model is applied
            debye_temperature (float): Debye temperature
            q_array (np.ndarray): q vectors

        Returns:
            Phonon: Debye phonon
        """
        phonon = Phonon(lattice, debye_temperature)

        # check the type and shape fo k_array
        if isinstance(q_array, list):
            q_array = np.array(q_array)

        if q_array.shape == (lattice.n_dim,):
            q_array = np.array([q_array])

        # set about q vectors
        phonon.q = q_array
        phonon.n_q = len(q_array)
        
        # set about values
        phonon.eigenenergies = phonon.calculate_eigenenergies(phonon.q)
        phonon.eigenvectors = phonon.calculate_eigenvectors(phonon.q)
        phonon.zero_point_lengths = phonon.calculate_zero_point_lengths()
        
        return phonon
    
    @classmethod
    def create_from_n(cls, lattice: Lattice, debye_temperature: float, n_q_array: np.ndarray | list[int]) -> 'Phonon':
        """Create Debye phonon from number of q vectors (array-type)

        Args:
            lattice (Lattice): Lattice on which the Debye model is applied
            debye_temperature (float): Debye temperature
            n_q_array (np.ndarray | list[int]): Number of q arrays

        Returns:
            Phonon: Debye phonon
        """
        phonon = Phonon(lattice, debye_temperature)
        
        # set about q vectors
        phonon.q = lattice.reciprocal.get_monkhorst_pack_grid(*n_q_array)
        phonon.n_q = np.prod(n_q_array)

        # set about values
        phonon.eigenenergies = phonon.calculate_eigenenergies(phonon.q)
        phonon.eigenvectors = phonon.calculate_eigenvectors(phonon.q)
        phonon.zero_point_lengths = phonon.calculate_zero_point_lengths()
        
        return phonon
    
    @classmethod
    def create_from_path(cls, lattice: Lattice, debye_temperature: float, q_path: PathValues) -> 'Phonon':
        """Create Debye phonon from q path

        Args:
            lattice (Lattice): Lattice on which the Debye model is applied
            debye_temperature (float): Debye model
            q_path (PathValues): q path

        Returns:
            Phonon: Debye phonon
        """
        phonon = Phonon.create_from_q(lattice, debye_temperature, q_path.values)
        
        return phonon

    def clone_with_q_grid(self, q_array: np.ndarray) -> 'Phonon':
        """Clone Debye phonon with changing q vectors

        Args:
            q_array (np.ndarray): q vectors

        Returns:
            Phonon: Debye phonon
        """
        phonon = self.create_from_q(self.lattice, self.debye_temperature, q_array)
        phonon.n_q = self.n_q
        
        return phonon

    def calculate_eigenenergies(self, q_array: np.ndarray) -> np.ndarray:
        """Calculate phonon eigenenergies.

        Args:
            q (np.ndarray): q vectors.

        Returns:
            np.ndarray: Phonon eigenenergies.
        """
        return self.speed_of_sound * np.linalg.norm(q_array, axis=-1)

    def calculate_eigenvectors(self, q_array: np.ndarray) -> np.ndarray:
        """Calculate phonon eigenvectors.

        Args:
            q (np.ndarray): q vectors

        Returns:
            np.ndarray: Phonon eigenvectors.
        """
        q_norm = np.repeat(np.linalg.norm(q_array, axis=-1, keepdims=True), q_array.shape[-1], axis=-1)
        
        return 1.0j * safe_divide(q_array, q_norm)

    def update(self, q: np.ndarray) -> None:
        """Update attributes

        Args:
            q (np.ndarray): q vectors
        """
        self.q = q
        self.eigenenergies = self.get_eigenenergies(q)
        self.eigenvectors = self.get_eigenvectors(q)
        self.zero_point_lengths = self.calculate_zero_point_lengths()

    def calculate_speed_of_sound(self) -> float:
        """Calculate the speed of sound.

        Returns:
            float: The speed of sound.
        """
        try:
            number_density = self.lattice.n_atoms / self.lattice.primitive.volume
        except ZeroDivisionError:
            ValueError("Lattice volume must be positive.")

        debye_frequency = self.debye_temperature * Energy.KELVIN["->"]

        if self.lattice.n_dim == 3:
            speed_of_sound = debye_frequency * (6.0 * np.pi ** 2 * number_density) ** (-1.0/3.0)
        elif self.lattice.n_dim == 2:
            speed_of_sound = debye_frequency * (4.0 * np.pi * number_density) ** (-1.0/2.0)
        elif self.lattice.n_dim == 1:
            speed_of_sound = debye_frequency / (2.0 * np.pi * number_density)
        
        return speed_of_sound
    
    def calculate_zero_point_lengths(self) -> np.ndarray:
        """Calculate zero point lengths

        Returns:
            np.ndarray: Zero point lengths
        """
        return safe_divide(1.0, np.sqrt(2.0 * self.lattice.mass * self.eigenenergies))


from unittest import TestCase
import numpy as np
from elphem.common.unit import Length
from elphem.lattice.lattice import Lattice3D
from elphem.phonon.phonon import Phonon

class TestUnit(TestCase):
    def setUp(self) -> None:
        a = 2.98 * Length.ANGSTROM["->"]
        self.debye_temperature = 344.0

        self.lattice = Lattice3D('bcc', 'Li', a)

    def test_create_from_q(self):
        q = np.zeros(3)
        phonon = Phonon.create_from_q(self.lattice, self.debye_temperature, q)

        omega = phonon.eigenenergies
        e = phonon.eigenvectors
        self.assertEqual(len(omega), phonon.n_q)
        self.assertEqual(len(e), phonon.n_q)

    def test_create_from_n(self):
        n_q_array = [8,8,8]
        phonon = Phonon.create_from_n(self.lattice, self.debye_temperature, n_q_array)

        omega = phonon.eigenenergies
        e = phonon.eigenvectors
        
        self.assertEqual(len(omega), phonon.n_q)
        self.assertEqual(len(e), phonon.n_q)

    def test_create_from_path(self):
        q_names = ["G", "H", "N", "G", "P", "H"]
        q_path = self.lattice.get_k_path(q_names, 20)
        
        phonon = Phonon.create_from_path(self.lattice, self.debye_temperature, q_path)
        
        omega = phonon.eigenenergies
        e = phonon.eigenvectors

        self.assertEqual(len(omega), phonon.n_q)
        self.assertEqual(len(e), phonon.n_q)


from unittest import TestCase
import numpy as np

from elphem.common.unit import Length
from elphem.lattice.lattice import *

class TestUnit(TestCase):
    def setUp(self) -> None:
        a = 2.98 * Length.ANGSTROM["->"]
        self.lattice = Lattice3D('bcc', 'Li', a)
        self.basis_primitive = self.lattice.primitive.basis
        self.basis_reciprocal = self.lattice.reciprocal.basis

    def test_vector(self):
        for b in [self.basis_primitive, self.basis_reciprocal]:
            self.assertEqual(b.shape, (3,3))

    def test_grid(self):
        n = [8, 8, 8]

        grid = self.lattice.reciprocal.get_monkhorst_pack_grid(*n)
        
        correct_shape = (np.prod(n), 3)
        
        self.assertEqual(grid.shape, correct_shape)


from unittest import TestCase
import numpy as np
from elphem.common.unit import Length
from elphem.lattice.lattice import Lattice3D
from elphem.electron.electron import Electron

class TestUnit(TestCase):
    def setUp(self) -> None:
        a = 2.98 * Length.ANGSTROM["->"]
        self.lattice = Lattice3D('bcc', 'Li', a)

    def test_create_from_n(self):
        electron = Electron.create_from_n(self.lattice, n_electrons=1, n_bands=20, n_k_array=[8,8,8])
        eigenenergies = electron.eigenenergies
        
        correct_shape = (len(electron.g), len(electron.k))
        self.assertEqual(eigenenergies.shape, correct_shape)

    def test_create_from_k(self):
        k = np.array([0., 0., 0.])
        electron = Electron.create_from_k(self.lattice, n_electrons=1, n_bands=20, k_array=k)
        eigenenergies = electron.eigenenergies
        
        correct_shape = (len(electron.g), len(electron.k))
        self.assertEqual(eigenenergies.shape, correct_shape)

    def test_create_from_path(self):
        k_names = ["G", "H", "N", "G", "P", "H"]
        k_path = self.lattice.get_k_path(k_names, 20)
        
        electron = Electron.create_from_path(self.lattice, n_electrons=1, n_bands=20, k_path=k_path)
        eigenenergies = electron.eigenenergies
        
        correct_shape = (len(electron.g), len(k_path.values))
        self.assertEqual(eigenenergies.shape, correct_shape)


import os
from dataclasses import dataclass

@dataclass
class ProgressBar:
    """Progress Bar
    
    Attributes:
        name: Name of processes.
        n_processes: Number of processes
        step: Step for printing the progress bar.
    """
    name: str
    n_processes: int
    step: int = 5

    def __post_init__(self):
        self.percentage = 0
        self.n_progress_tile = 0

    def print(self, count: int) -> None:
        """Print a progress bar.

        Args:
            count (int): Count for processes.
        """
        self.percentage = int(count / self.n_processes * 100)
        n_progress_tile = int(self.percentage / self.step)

        if n_progress_tile > self.n_progress_tile:
            os.system('clear')
            print('Process: {}'.format(self.name))
            print('{} %: '.format(self.percentage) + '█' * n_progress_tile)
            self.n_progress_tile = n_progress_tile


from dataclasses import dataclass
import numpy as np

@dataclass
class PathValues:
    """A class for values with path.
    
    Attributes:
        major_scales (np.ndarray): A numpy array of major scales of path.
        minor_scales (np.ndarray): A numpy array of minor scales of path.
        values (np.ndarray): A numpy array of values.
    """
    major_scales: np.ndarray
    minor_scales: np.ndarray
    values: np.ndarray


from dataclasses import dataclass
import numpy as np

from elphem.common.atomic_weight import AtomicWeight
from elphem.common.unit import Mass
from elphem.lattice.primitive_cell import PrimitiveCell3D, PrimitiveCell2D, PrimitiveCell1D
from elphem.lattice.reciprocal_cell import ReciprocalCell3D, ReciprocalCell2D, ReciprocalCell1D
from elphem.lattice.lattice_constant import LatticeConstant3D, LatticeConstant2D, LatticeConstant1D
from elphem.lattice.path import PathValues

@dataclass
class Lattice:
    """A class to simulate an empty lattice for a given crystal structure and lattice constant a.
    Attributes:
        crystal_structure (str): A string for crystal structures
        atoms (str | list[str] | np.ndarray): Name(s) of atoms
        n_atoms (int): Number of atoms
        a (float): A lattice constant
        primitive (PrimitiveCell): A primitive cell
        reciprocal (ReciprocalCell): A reciprocal cell
        mass (float): Mass of the empty lattice
    """
    crystal_structure: str
    atoms: str | list[str] | np.ndarray
    a: float
    primitive = None
    reciprocal = None

    def __post_init__(self):
        self.correct_atoms()
        self.set_about_atoms()

    def set_about_atoms(self) -> None:
        """Set number of atoms and mass.
        """
        self.n_atoms = len(self.atoms)

        masses = np.array(AtomicWeight.get_from_list(self.atoms)) * Mass.DALTON["->"]
        self.mass = np.average(masses)

    def correct_atoms(self) -> None:
        """Correct the type of atoms
        """
        if isinstance(self.atoms, str):
            self.atoms = [self.atoms]
        elif isinstance(self.atoms, np.ndarray):
            self.atoms = self.atoms.tolist()
    
    def get_k_path(self, k_names: list[str], n_split: int) -> PathValues:
        """Get a path in the reciprocal space

        Args:
            k_names (list[str]): A list for names of special k points.
            n_split (int): Number of points between each special point.

        Returns:
            PathValues: Values with path
        """
        return self.reciprocal.get_path(k_names, n_split)

class Lattice3D(Lattice):
    """A class for 3D lattice"""
    def __init__(self, crystal_structure: str, atoms: str | list[str] | np.ndarray, a: float):
        super().__init__(crystal_structure, atoms, a)
        self.n_dim = 3
        self.set_constants()

        self.primitive = PrimitiveCell3D(self.constants)
        self.reciprocal = ReciprocalCell3D(self.constants)

    def set_constants(self) -> None:
        """Determines lattice constants based on the crystal structure.
        
        Raises:
            ValueError: If an invalid crystal structure name is specified.
        """
        crystal_structure_lower = self.crystal_structure.lower()

        alpha_values = {
            'bcc': 109.47,
            'fcc': 60.0,
            'sc': 90.0
        }

        alpha = alpha_values.get(crystal_structure_lower)

        if alpha is not None:
            self.constants = LatticeConstant3D(self.a, self.a, self.a, alpha, alpha, alpha, crystal_structure_lower)
        else:
            raise ValueError("Invalid crystal structure specified.")

class Lattice2D(Lattice):
    """A class for 2D lattice"""
    def __init__(self, crystal_structure: str, atoms: str | list[str] | np.ndarray, a: float):
        super().__init__(crystal_structure, atoms, a)
        self.n_dim = 2
        self.set_constants()

        self.primitive = PrimitiveCell2D(self.constants)
        self.reciprocal = ReciprocalCell2D(self.constants)

    def set_constants(self) -> None:
        """Determines lattice constants based on the crystal structure.
        
        Raises:
            ValueError: If an invalid crystal structure name is specified.
        """
        crystal_structure_lower = self.crystal_structure.lower()

        alpha_values = {
            'square': 90.0,
            'hexagonal': 120.0
        }

        alpha = alpha_values.get(crystal_structure_lower)

        if alpha is not None:
            self.constants = LatticeConstant2D(self.a, self.a, alpha, crystal_structure_lower)
        else:
            raise ValueError("Invalid crystal structure specified.")

class Lattice1D(Lattice):
    """A class for 1D lattice"""
    def __init__(self, atoms: str | list[str] | np.ndarray, a: float):
        super().__init__('', atoms, a)
        self.n_dim = 1
        self.set_constants()

        self.primitive = PrimitiveCell1D(self.constants)
        self.reciprocal = ReciprocalCell1D(self.constants)

    def set_constants(self) -> None:
        """Determines lattice constants based on the crystal structure.
        
        Raises:
            ValueError: If an invalid crystal structure name is specified.
        """

        self.constants = LatticeConstant1D(self.a)


import numpy as np
from scipy.spatial.transform import Rotation

class LatticeRotation:
    @staticmethod
    def normalize(v: np.ndarray) -> np.ndarray:
        """Normalize the given vector.

        Args:
            v (np.ndarray): The vector to normalize.

        Returns:
            np.ndarray: The normalized vector.

        Raises:
            ValueError: If the input vector is zero.
        """
        v_norm = np.linalg.norm(v)

        if v_norm == 0.0:
            return v

        return v / v_norm

class LatticeRotation3D(LatticeRotation):
    """Provides methods to optimize and adjust 3D lattice orientations through rotation operations."""
    
    @classmethod
    def calculate_optimized_basis(cls, basis: np.ndarray, axis: np.ndarray) -> np.ndarray:
        """Optimize the rotation of the given basis to align with the specified axis.

        Args:
            basis (np.ndarray): The basis vectors of the lattice to be rotated.
            axis (np.ndarray): The target axis for alignment.

        Returns:
            np.ndarray: The optimized basis vectors aligned as closely as possible to the target axis.

        Raises:
            ValueError: If the axis vector is zero.
        """
        basis_rotated = cls.search_optimized_direction(basis, axis)
        basis_rotated = cls.search_optimized_posture(basis_rotated, axis)
        return basis_rotated

    @classmethod
    def search_optimized_direction(cls, basis: np.ndarray, axis: np.ndarray) -> np.ndarray:
        """search_optimized_direction the given basis direction to the specified axis using quaternion rotation.

        Args:
            basis (np.ndarray): The initial basis vectors.
            axis (np.ndarray): The axis to align the basis vectors to.

        Returns:
            np.ndarray: The basis vectors rotated to search_optimized_direction the direction of the axis.

        Raises:
            ValueError: If the axis vector is zero.
        """
        if np.linalg.norm(axis) == 0.0:
            raise ValueError("Axis vector should not be zero.")
        
        direction = cls.normalize(basis[0] + basis[1] + basis[2])
        n = cls.normalize(axis)

        cross = cls.normalize(np.cross(direction, n))
        dot = np.dot(direction, n)
        theta = np.arccos(dot) / 2.0
        sin_theta = np.sin(theta)

        quaternion = np.array([
            cross[0] * sin_theta,
            cross[1] * sin_theta,
            cross[2] * sin_theta,
            np.cos(theta)
        ])

        rotation = Rotation.from_quat(quaternion)
        basis_rotated = rotation.apply(basis)
        
        return basis_rotated

    @classmethod
    def rotate_vector_around_axis(cls, axis: np.ndarray, v: np.ndarray, theta: float) -> np.ndarray:
        """Rotate vector v around the given axis by angle theta.

        Args:
            axis (np.ndarray): The axis of rotation.
            v (np.ndarray): The vector to rotate.
            theta (float): The angle in radians to rotate the vector.

        Returns:
            np.ndarray: The rotated vector.

        Raises:
            ValueError: If the axis vector is zero.
        """
        if np.linalg.norm(axis) == 0.0:
            raise ValueError("Axis vector should not be zero.")
        
        axis_dot_v = np.dot(axis, v)
        cos_theta = np.cos(theta)
        v_rotated = cos_theta * v
        v_rotated += np.sin(theta) * np.cross(axis, v)
        v_rotated += (1.0 - cos_theta) * axis_dot_v * axis

        return v_rotated

    @classmethod
    def search_optimized_posture(cls, basis: np.ndarray, axis: np.ndarray, angle_max: float = 360.0) -> np.ndarray:
        """Search for the best posture of the basis aligned with the given axis.

        Args:
            basis (np.ndarray): The basis to be rotated.
            axis (np.ndarray): The axis to align with.
            angle_max (float): The maximum angle range to search in degrees.

        Returns:
            np.ndarray: The optimally aligned basis after posture search.

        Raises:
            ValueError: If the axis vector is zero.
        """
        if np.linalg.norm(axis) == 0.0:
            raise ValueError("Axis vector should not be zero.")
        
        angle = np.linspace(0.0, np.radians(angle_max), 1000)
        s_min = 1.0e+100
        basis_searched = basis
        
        n = np.zeros(basis.shape)
        u = np.identity(3)
                    
        axis_normalized = cls.normalize(axis)
        for theta in angle:
            basis_rotated = cls.rotate_vector_around_axis(axis_normalized, basis, theta)
            
            s = 0.0
            for i in range(3):
                j = i
                k = (i+1) % 3
                n[i] = np.cross(basis_rotated[j], basis_rotated[k])
                s += np.dot(n[i], u[i])
            
            if s < s_min:
                s_min = s
                basis_searched = basis_rotated

        return basis_searched

class LatticeRotation2D(LatticeRotation):
    """Provides methods to optimize and adjust 2D lattice orientations through rotation operations."""
    
    @classmethod
    def calculate_optimized_basis(cls, basis: np.ndarray, axis: np.ndarray) -> np.ndarray:
        """Optimize the rotation of the given basis to align with the specified axis.

        Args:
            basis (np.ndarray): The basis vectors of the lattice to be rotated.
            axis (np.ndarray): The target axis for alignment.

        Returns:
            np.ndarray: The optimized basis vectors aligned as closely as possible to the target axis.

        Raises:
            ValueError: If the axis vector is zero.
        """
        basis_rotated = cls.search_optimized_direction(basis, axis)
        return basis_rotated

    @classmethod
    def search_optimized_direction(cls, basis: np.ndarray, axis: np.ndarray) -> np.ndarray:
        """search_optimized_direction the given basis direction to the specified axis using quaternion rotation.

        Args:
            basis (np.ndarray): The initial basis vectors.
            axis (np.ndarray): The axis to align the basis vectors to.

        Returns:
            np.ndarray: The basis vectors rotated to search_optimized_direction the direction of the axis.

        Raises:
            ValueError: If the axis vector is zero.
        """
        if np.linalg.norm(axis) == 0.0:
            raise ValueError("Axis vector should not be zero.")
        
        direction = cls.normalize(basis[0] + basis[1])
        n = cls.normalize(axis)

        dot = np.dot(direction, n)

        theta = np.arccos(dot)
        
        sin_theta = np.sin(theta)
        cos_theta = np.cos(theta)
        
        rotation_matrix = np.array([
            [cos_theta, -sin_theta],
            [sin_theta, cos_theta]
        ])

        basis_rotated = basis @ rotation_matrix
        
        return basis_rotated


import sys
import numpy as np

from elphem.common.unit import Energy
from elphem.common.function import safe_divide

def boltzmann_distribution(temperature: float, energy: float | np.ndarray) -> float | np.ndarray:
    """
    Calculates the occupation number of particles following the Boltzmann distribution.

    Args:
        temperature (float): Temperature in Kelvin.
        energy (float | np.ndarray): Energy value(s) in Hartree atomic units.

    Returns:
        float | np.ndarray: Occupation number(s) based on the Boltzmann distribution.
    """

    # Setting up system-related constants and warning filters
    float_min = sys.float_info.min
    float_max = sys.float_info.max

    kbt = max(temperature * Energy.KELVIN["->"], float_min)
    beta = safe_divide(1.0, kbt, default=float_max)

    ln = - beta * energy
    
    safe_ln_min = np.log(float_min)
    safe_ln_max = np.log(float_max)
    
    return np.exp(ln, out=np.zeros_like(energy), where=((ln > safe_ln_min) & (ln < safe_ln_max)))

def fermi_distribution(temperature: float, energy: float | np.ndarray) -> float | np.ndarray:
    """
    Calculates the occupation number of particles following the Fermi-Dirac distribution.

    Args:
        temperature (float): Temperature in Kelvin.
        energy (float | np.ndarray): Energy value(s) in Hartree atomic units.

    Returns:
        float | np.ndarray: Occupation number(s) based on the Fermi-Dirac distribution.
    """
    boltzmann_factor = boltzmann_distribution(temperature, energy)
    inv_boltzmann_factor = safe_divide(1.0, boltzmann_factor)

    return safe_divide(1.0, inv_boltzmann_factor + 1.0)

def bose_distribution(temperature: float, energy: float | np.ndarray) -> float | np.ndarray:
    """
    Calculates the occupation number of particles following the Bose-Einstein distribution.

    Args:
        temperature (float): Temperature in Kelvin.
        energy (float | np.ndarray): Energy value(s) in Hartree atomic units.

    Returns:
        float | np.ndarray: Occupation number(s) based on the Bose-Einstein distribution.
    """
    boltzmann_factor = boltzmann_distribution(temperature, energy)
    inv_boltzmann_factor = safe_divide(1.0, boltzmann_factor)
    
    return safe_divide(1.0, inv_boltzmann_factor - 1.0)

def gaussian_distribution(sigma: float, energy: float | np.ndarray) -> float | np.ndarray:
    """
    Calculates the probability density of particles following the Gaussian distribution.

    Args:
        sigma (float): Standard deviation of the Gaussian distribution.
        energy (float | np.ndarray): Energy value(s) to evaluate the Gaussian function.

    Returns:
        float | np.ndarray: Probability density(s) based on the Gaussian distribution.
    """
    
    return np.exp(- energy ** 2 / (2.0 * sigma ** 2)) / (np.sqrt(2.0 * np.pi) * sigma)


from unittest import TestCase
from elphem.common.brillouin import SpecialPoints3D, SpecialPoints2D, SpecialPoints1D

class TestUnit(TestCase):
    def test_3D_points(self):
        gamma = SpecialPoints3D.Gamma["G"]
        fcc_gamma = SpecialPoints3D.FCC["G"]
        bcc_h = SpecialPoints3D.BCC["H"]
        hexagonal_h = SpecialPoints3D.Hexagonal["H"]

        self.assertEqual(gamma, fcc_gamma)
        self.assertNotEqual(bcc_h, hexagonal_h)
    
    def test_2D_points(self):
        gamma = SpecialPoints2D.Gamma["G"]
        square_gamma = SpecialPoints2D.Square["G"]
        square_m = SpecialPoints2D.Square["M"]
        hexagonal_m = SpecialPoints2D.Hexagonal["M"]

        self.assertEqual(gamma, square_gamma)
        self.assertNotEqual(square_m, hexagonal_m)

    def test_1D_points(self):
        x = SpecialPoints1D.Line["X"]

        self.assertEqual(x, 0.5)


import math
import numpy as np

from elphem.lattice.lattice import Lattice
from elphem.lattice.path import PathValues

class Electron:
    """Represents free electrons in a given empty lattice.
    
    Attributes:
        lattice (Lattice): Lattice on which the free electron model is applied.
        n_electrons (int): Number of electrons per unit cell.
        n_bands (int): Number of energy bands considered.
        n_k (int): Number of k_points.
        g (np.ndarray): Reciprocal lattice vectors.
        k (np.ndarray): Vectors in the reciprocal space.
        eigenenergies (np.ndarray): Eigenenergies of free electron model.
        fermi_energy (float): Fermi energy
        fermi_wave_number (float): Fermi wave-number
        thomas_fermi_wave_number (float): Thomas-Fermi wave-number
    """
    
    def __init__(self, lattice: Lattice, n_electrons: int):
        """
        Args:
            lattice (Lattice): Lattice on which the free electron model is applied.
            n_electrons (int): Number of electrons per unit cell.
        """
        
        self.lattice = lattice
        self.n_electrons = n_electrons
        self.n_bands = None
        self.n_k = None
        self.g = None
        self.k = None
        self.eigenenergies = None

        self.fermi_energy = self.calculate_fermi_energy()
        self.fermi_wave_number = self.calculate_fermi_wave_number()
        self.thomas_fermi_wave_number = self.calculate_thomas_fermi_wave_number()
        
    @classmethod
    def create_from_n(cls, lattice: Lattice, n_electrons: int, n_bands: int, n_k_array: np.ndarray) -> 'Electron':
        """Create free electrons from the number of the number of k (array-type).

        Args:
            lattice (Lattice): Lattice on which the free electron model is applied
            n_electrons (int): Number of electrons
            n_bands (int): Number of electron bands
            n_k_array (np.ndarray): Number of k vectors (array-type)

        Returns:
            Electron: Free electron
        """
        electron = Electron(lattice, n_electrons)

        # set about k vectors
        electron.k = lattice.reciprocal.get_monkhorst_pack_grid(*n_k_array)
        electron.n_k = np.prod(n_k_array)

        # update values
        electron.update_band(n_bands)
        electron.update_eigenenergies()
        
        return electron

    @classmethod
    def create_from_k(cls, lattice: Lattice, n_electrons: int, n_bands: int, k_array: np.ndarray) -> 'Electron':
        """Create free electrons from k vectors.

        Args:
            lattice (Lattice): Lattice on which the free electron model is applied
            n_electrons (int): Number of electrons
            n_bands (int): Number of electron bands
            k_array (np.ndarray): k vectors

        Returns:
            Electron: Free electron
        """
        electron = Electron(lattice, n_electrons)

        # check the type and shape fo k_array
        if isinstance(k_array, list):
            k_array = np.array(k_array)

        if k_array.shape == (lattice.n_dim,):
            k_array = np.array([k_array])

        # set about k vectors
        electron.n_k = len(k_array)
        electron.k = k_array

        # update values
        electron.update_band(n_bands)
        electron.update_eigenenergies()
        
        return electron

    @classmethod
    def create_from_gk_grid(cls, lattice: Lattice, n_electrons: int, g_array: np.ndarray, k_array: np.ndarray) -> 'Electron':
        """Create free electrons from G-k grid.

        Args:
            lattice (Lattice): Lattice on which the free electron model is applied
            n_electrons (int): Number of electrons
            g_array (np.ndarray): Reciprocal lattice vectors
            k_array (np.ndarray): k vectors

        Returns:
            Electron: Free electron
        """
        electron = Electron(lattice, n_electrons)
        
        electron.g = g_array
        electron.k = k_array
        
        electron.update_eigenenergies(expand_g=False)
        
        return electron
    
    @classmethod
    def create_from_path(cls, lattice: Lattice, n_electrons: int, n_bands: int, k_path: PathValues) -> 'Electron':
        """Create free electrons from k-path.

        Args:
            lattice (Lattice): Lattice on which the free electron model is applied
            n_electrons (int): Number of electrons
            n_bands (int): Number of bands
            k_path (PathValues): Path in the reciprocal lattice

        Returns:
            Electron: Free electron
        """
        electron = Electron.create_from_k(lattice, n_electrons, n_bands, k_path.values)
        
        return electron

    def clone_with_gk_grid(self, g_array: np.ndarray, k_array: np.ndarray) -> 'Electron':
        """Clone a free electron with changing G and k vectors.

        Args:
            g_array (np.ndarray): Reciprocal lattice vectors
            k_array (np.ndarray): k vectors

        Returns:
            Electron: Free electron
        """
        electron = self.create_from_gk_grid(self.lattice, self.n_electrons, g_array, k_array)

        # set about numbers
        electron.n_k = self.n_k
        electron.n_bands = self.n_bands
        
        return electron

    def calculate_eigenenergies(self, k_array: np.ndarray, g_array: np.ndarray = None) -> np.ndarray:
        """Calculate the electron eigenenergies at wave vector k.
        
        Args:
            k_array (np.ndarray): k vectors
            g_array (np.ndarray): Reciprocal lattice vectors

        Returns:
            np.ndarray: The electron eigenenergies
        """
        if g_array is None:
            eigenenergies =  0.5 * np.linalg.norm(k_array, axis=-1) ** 2 - self.fermi_energy
        else:
            eigenenergies = np.array([0.5 * np.linalg.norm(k_array + g, axis=-1) ** 2 for g in g_array])
            eigenenergies -= self.fermi_energy
        
        return eigenenergies

    def calculate_dos(self, omega: float | np.ndarray) -> np.ndarray:
        """Calculate DOS by using analytical forms.

        Args:
            omega (float | np.ndarray): frequencies

        Returns:
            np.ndarray: DOS
        """
        omega_plus_fermi_energy = omega + self.fermi_energy
        if self.lattice.n_dim == 3:
            coefficient = 8.0 * np.pi / self.lattice.reciprocal.volume
            return coefficient * np.sqrt(2.0 * omega_plus_fermi_energy)
        elif self.lattice.n_dim == 2:
            coefficient = 4.0 * np.pi / self.lattice.reciprocal.volume
            return coefficient
        else:
            coefficient = 2.0 / self.lattice.reciprocal.volume
            return coefficient / np.sqrt(2.0 * omega_plus_fermi_energy)

    def update_band(self, n_bands: int) -> None:
        """Update attributes about bands

        Args:
            n_bands (int): Number of bands
        """
        self.n_bands = n_bands
        self.g = self.lattice.reciprocal.get_reciprocal_vectors(self.n_bands)

    def update_eigenenergies(self, expand_g: bool = True) -> None:
        """Update eigenenergies

        Args:
            expand_g (bool, optional): add axis about reciprocal lattice vectors. Defaults to True.
        """
        if expand_g:
            self.eigenenergies = self.calculate_eigenenergies(self.k, self.g)
        else:
            self.eigenenergies = self.calculate_eigenenergies(self.k + self.g)

    def calculate_fermi_energy(self) -> float:
        """Calculate the Fermi energy.

        Returns:
            float: The Fermi energy.
        """
        gamma = math.gamma(self.lattice.n_dim / 2.0 + 1.0)
        coefficient = 2.0 * np.pi
        electron_density = self.n_electrons / self.lattice.primitive.volume
        
        return coefficient * (0.5 * gamma * electron_density) ** (2.0 / self.lattice.n_dim)
    
    def calculate_fermi_wave_number(self) -> float:
        """Calculate the Fermi wave-number from Fermi energy.

        Returns:
            float: The Fermi wave-number
        """
        
        return np.sqrt(2.0 * self.fermi_energy)
    
    def calculate_thomas_fermi_wave_number(self) -> float:
        """Calculate the Thomas-Fermi wave-number from Fermi wave-number

        Returns:
            float: The Thomas-Fermi wave-number
        """
        
        coefficient = 4.0 / np.pi * (4.0 / (9.0 * np.pi)) ** (1.0 / 3.0)
        r = (self.lattice.primitive.volume * math.gamma(self.lattice.n_dim / 2.0 + 1.0) / (self.n_electrons * np.pi ** (self.lattice.n_dim / 2.0))) ** (1.0 / self.lattice.n_dim)
        
        return coefficient * r * self.fermi_wave_number ** 2


from setuptools import setup, find_packages

DESCRIPTION = 'Elphem: Calculating electron-phonon interactions with the empty lattice.'
NAME = 'elphem'
AUTHOR = 'Kohei Ishii'
AUTHOR_EMAIL = ''
URL = 'https://github.com/cohsh/elphem'
LICENSE = 'MIT'
DOWNLOAD_URL = URL
VERSION = '0.3.3'
PYTHON_REQUIRES = '>=3.10'
INSTALL_REQUIRES = ['numpy', 'scipy']

CLASSIFIERS=[
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python',
    'Programming Language :: Python :: 3.10',
    'Programming Language :: Python :: 3.11',
    'Programming Language :: Python :: 3.12'
]

try:
    with open('README.md', 'r', encoding='utf-8') as fp:
        LONG_DESCRIPTION = fp.read()
except FileExistsError:
    LONG_DESCRIPTION = DESCRIPTION

LONG_DESCRIPTION_CONTENT_TYPE = 'text/markdown'

setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=AUTHOR,
    maintainer_email=AUTHOR_EMAIL,
    url=URL,
    download_url=URL,
    packages=find_packages(),
    classifiers=CLASSIFIERS,
    license=LICENSE,
    install_requires=INSTALL_REQUIRES
)


import numpy as np
from dataclasses import dataclass

@dataclass
class LatticeConstant3D:
    """Defines the lattice constants and angles for a 3D crystal structure.

    Attributes:
        a (float): Length of the first lattice vector.
        b (float): Length of the second lattice vector.
        c (float): Length of the third lattice vector.
        alpha (float): Angle between b and c lattice vectors.
        beta (float): Angle between a and c lattice vectors.
        gamma (float): Angle between a and b lattice vectors.
        crystal_structure (str): The type of crystal structure (e.g., 'bcc', 'fcc', 'sc').
    """
    a: float
    b: float
    c: float
    alpha: float
    beta: float
    gamma: float
    crystal_structure: str
    
    def __post_init__(self):
        """Converts angles to radians and stores lengths and angles as numpy arrays."""
        self.length = np.array([self.a, self.b, self.c])
        self.angle = np.radians(np.array([self.alpha, self.beta, self.gamma]))

@dataclass
class LatticeConstant2D:
    """Defines the lattice constants and angles for a 2D crystal structure.

    Attributes:
        a (float): Length of the first lattice vector.
        b (float): Length of the second lattice vector.
        alpha (float): Angle between a and b lattice vectors.
        crystal_structure (str): The type of crystal structure (e.g., 'square', 'hexagonal').
    """
    a: float
    b: float
    alpha: float
    crystal_structure: str
    
    def __post_init__(self):
        """Converts angles to radians and stores lengths as numpy arrays."""
        self.length = np.array([self.a, self.b])
        self.angle = np.radians(self.alpha)

@dataclass
class LatticeConstant1D:
    """Defines the lattice constants and angles for a crystal structure.

    Attributes:
        a (float): Length of the lattice vector.
    """
    a: float
    
    def __post_init__(self):
        """Stores the length."""
        self.length = self.a